from typing import List

# When updating this file, please also update the docs:
# https://docs.v7labs.com/docs/import-1
supported_formats: List[str] = [
    "coco",
    "dataloop",
    "csv_tags",
    "csv_tags_video",
    "darwin",
    "labelbox",
    "pascal_voc",
    "superannotate",
    "nifti",
]
